#!/usr/local/bin/python
# -*- coding: utf-8 -*-
"""
This module implements a Description Horn Logic implementation as defined
by Grosof, B. et.al. ("Description Logic Programs: Combining Logic Programs with 
Description Logic" [1]) in section 4.4.  As such, it implements recursive mapping
functions "T", "Th" and "Tb" which result in "custom" (dynamic) rulesets, RIF Basic 
Logic Dialect: Horn rulesets [2], [3].  The rulesets are evaluated against an 
efficient RETE-UL network.

It is a Description Logic Programming [1] Implementation on top of RETE-UL:

"A DLP is directly defined as the LP-correspondent of a def-Horn
ruleset that results from applying the mapping T ."

The mapping is as follows:

== Core (Description Horn Logic) ==

== Class Equivalence ==

T(owl:equivalentClass(C,D)) -> { T(rdfs:subClassOf(C,D) 
                                 T(rdfs:subClassOf(D,C) }
                                 
== Domain and Range Axioms (Base Description Logic: "ALC") ==                                                                                                       

T(rdfs:range(P,D))  -> D(y) := P(x,y)
T(rdfs:domain(P,D)) -> D(x) := P(x,y)

== Property Axioms (Role constructors: "I") ==

T(rdfs:subPropertyOf(P,Q))     -> Q(x,y) :- P(x,y)
T(owl:equivalentProperty(P,Q)) -> { Q(x,y) :- P(x,y)
                                    P(x,y) :- Q(x,y) }
T(owl:inverseOf(P,Q))          -> { Q(x,y) :- P(y,x)
                                    P(y,x) :- Q(x,y) }
T(owl:TransitiveProperty(P))   -> P(x,z) :- P(x,y) ^ P(y,z)                                                                        

[1] http://www.cs.man.ac.uk/~horrocks/Publications/download/2003/p117-grosof.pdf
[2] http://www.w3.org/2005/rules/wg/wiki/Core/Positive_Conditions
[3] http://www.w3.org/2005/rules/wg/wiki/asn06

"""

from __future__ import generators
from sets import Set
from rdflib import BNode, RDF, Namespace, Variable, RDFS
from rdflib.util import first
from rdflib.Collection import Collection
from rdflib.store import Store,VALID_STORE, CORRUPTED_STORE, NO_STORE, UNKNOWN
from rdflib import Literal, URIRef
from pprint import pprint, pformat
import sys, copy
from rdflib.Graph import QuotedGraph, Graph
from rdflib.store.REGEXMatching import REGEXTerm, NATIVE_REGEX, PYTHON_REGEX
from FuXi.Rete.RuleStore import Formula
from FuXi.Rete.AlphaNode import AlphaNode
from FuXi.Horn.PositiveConditions import And, Or, Uniterm, Condition, Atomic,SetOperator,Exists
from FuXi.Horn.HornRules import Clause as OriginalClause
from FuXi.Rete.Util import renderNetwork
from cStringIO import StringIO

SKOLEMIZED_CLASS_NS=Namespace('http://code.google.com/p/python-dlp/wiki/SkolemTerm#')

non_DHL_OWL_Semantics=\
"""
@prefix log: <http://www.w3.org/2000/10/swap/log#>.
@prefix math: <http://www.w3.org/2000/10/swap/math#>.
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix : <http://eulersharp.sourceforge.net/2003/03swap/owl-rules#>.
@prefix list: <http://www.w3.org/2000/10/swap/list#>.
#Additional OWL-compliant semantics, mappable to Production Rules 

#Subsumption (purely for TBOX classification)
{?C rdfs:subClassOf ?SC. ?A rdfs:subClassOf ?C} => {?A rdfs:subClassOf ?SC}.
{?C owl:equivalentClass ?A} => {?C rdfs:subClassOf ?A. ?A rdfs:subClassOf ?C}.
{?C rdfs:subClassOf ?SC. ?SC rdfs:subClassOf ?C} => {?C owl:equivalentClass ?SC}.

{?C owl:disjointWith ?B. ?M a ?C. ?Y a ?B } => {?M owl:differentFrom ?Y}.

{?P owl:inverseOf ?Q. ?P a owl:InverseFunctionalProperty} => {?Q a owl:FunctionalProperty}.
{?P owl:inverseOf ?Q. ?P a owl:FunctionalProperty} => {?Q a owl:InverseFunctionalProperty}.

#Inverse functional semantics
{?P a owl:FunctionalProperty. ?S ?P ?O. ?S ?P ?Y} => {?O = ?Y}.
{?P a owl:InverseFunctionalProperty. ?S ?P ?O. ?Y ?P ?O} => {?S = ?Y}.
{?T1 = ?T2. ?S = ?T1} => {?S = ?T2}.
{?T1 ?P ?O. ?T1 = ?T2.} => {?T2 ?P ?O}.

#For OWL/InverseFunctionalProperty/premises004
{?C owl:oneOf ?L. ?L rdf:first ?X; rdf:rest rdf:nil. ?P rdfs:domain ?C} => {?P a owl:InverseFunctionalProperty}.
#For OWL/InverseFunctionalProperty/premises004
{?C owl:oneOf ?L. ?L rdf:first ?X; rdf:rest rdf:nil. ?P rdfs:range ?C} => {?P a owl:FunctionalProperty}.

{?S owl:differentFrom ?O} => {?O owl:differentFrom ?S}.
{?S owl:complementOf ?O} => {?O owl:complementOf ?S}.
{?S owl:disjointWith ?O} => {?O owl:disjointWith ?S}.

"""

NOMINAL_SEMANTICS=\
"""
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix list: <http://www.w3.org/2000/10/swap/list#>.

#For OWL/oneOf
{?C owl:oneOf ?L. ?X list:in ?L} => {?X a ?C}.
{?L rdf:first ?I} => {?I list:in ?L}.
{?L rdf:rest ?R. ?I list:in ?R} => {?I list:in ?L}.
"""

OWL_NS    = Namespace("http://www.w3.org/2002/07/owl#")

LOG = Namespace("http://www.w3.org/2000/10/swap/log#")
Any = None

LHS = 0
RHS = 1

def reduceAnd(left,right):
    if isinstance(left,And):
        left = reduce(reduceAnd,left)
    elif isinstance(right,And):
        right = reduce(reduceAnd,right)
    if isinstance(left,list) and not isinstance(right,list):
        return left+[right]
    elif isinstance(left,list) and isinstance(right,list):
        return left+right
    elif isinstance(left,list) and not isinstance(right,list):
        return left+[right]
    elif not isinstance(left,list) and isinstance(right,list):
        return [left]+right
    else:
        return [left,right]
    
def NormalizeClause(clause):
    def fetchFirst(gen):
        return first(gen)
    if hasattr(clause.head,'next') and not isinstance(clause.head,Condition):
        clause.head = fetchFirst(clause.head)
    if hasattr(clause.body,'next') and not isinstance(clause.body,Condition):
        clause.body = fetchFirst(clause.body)
#    assert isinstance(clause.head,(Atomic,And,Clause)),repr(clause.head)
#    assert isinstance(clause.body,Condition),repr(clause.body)
    if isinstance(clause.head,And):
        clause.head.formulae = reduce(reduceAnd,clause.head)
    if isinstance(clause.body,And):
        clause.body.formulae = reduce(reduceAnd,clause.body)
#    print "Normalized clause: ", clause
    return clause

class Clause(OriginalClause):
    """
    The RETE-UL algorithm supports conjunctions of facts in the head of a rule
    i.e.:   H1 ^ H2 ^ ... ^ H3 :- B1 ^  ^ Bm
    The Clause definition is overridden to permit this syntax (not allowed
    in definite LP or Horn rules)
    
    In addition, since we allow (in definite Horn) entailments beyond simple facts
    we ease restrictions on the form of the head to include Clauses
        """
    def __init__(self,body,head):
        self.body = body
        self.head = head
        if isinstance(head,Uniterm):
            from FuXi.Rete.Network import HashablePatternList
            try:
                antHash=HashablePatternList([term.toRDFTuple() 
                                    for term in body],skipBNodes=True)
                consHash=HashablePatternList([term.toRDFTuple() 
                                    for term in head],skipBNodes=True)                                                                                            
                self._hash = hash(antHash) ^ hash(consHash)
            except:
                self._hash = None
        else:
            self._hash = None

    def __hash__(self):
        if self._hash is None:
            from FuXi.Rete.Network import HashablePatternList
            antHash=HashablePatternList([term.toRDFTuple() 
                                for term in self.body],skipBNodes=True)
            consHash=HashablePatternList([term.toRDFTuple() 
                                for term in self.head],skipBNodes=True)                                                                                            
            self._hash = hash(antHash) ^ hash(consHash)
        return self._hash
                
    def __repr__(self):
        return "%r :- %r"%(self.head,self.body)

    def n3(self):
        return u'{ %s } => { %s }'%(self.body.n3(),self.head.n3())    

def makeRule(clause,nsMap):
    from FuXi.Horn.HornRules import Rule
    vars=set()
    for child in clause.head:
        if isinstance(child,Or):
            #Disjunction in the head, skip this rule:
            #When a disjunction occurs on the r.h.s. of a subclass axiom it 
            #becomes a disjunction in the head of the corresponding rule, and 
            #this cannot be handled within the def-Horn framework.            
            return None
        assert isinstance(child,Uniterm),repr(child)
        vars.update([term for term in child.toRDFTuple() if isinstance(term,Variable)])
    for child in clause.body:
        assert isinstance(child,Uniterm),repr(child)
        vars.update([term for term in child.toRDFTuple() if isinstance(term,Variable)])
    return Rule(clause,declare=vars,nsMapping=nsMap)

def MapDLPtoNetwork(network,factGraph,complementExpansions=[],constructNetwork=False):
    ruleset=[]
    for horn_clause in T(factGraph,complementExpansions=complementExpansions):
#        print "## RIF BLD Horn Rules: Before LloydTopor: ##\n",horn_clause
#        print "## RIF BLD Horn Rules: After LloydTopor: ##"
        for tx_horn_clause in LloydToporTransformation(horn_clause):
            tx_horn_clause = NormalizeClause(tx_horn_clause)
#            print tx_horn_clause
            disj = [i for i in breadth_first(tx_horn_clause.body) if isinstance(i,Or)]
            import warnings
            if len(disj)>1:
                raise
                warnings.warn("No support for multiple disjunctions in the body:\n"+repr(tx_horn_clause),UserWarning,1)
            elif disj:
                #Disjunctions in the body
#                print "Disjunction in the body!"
#                print tx_horn_clause
                disj = disj[0]
                for item in disj:
#                    print "\tDisjunction operand: ", item
                    #replace disj with item in tx_horn_clause.body
                    list(breadth_first_replace(tx_horn_clause.body,candidate=disj,replacement=item))
                    #Then we want to clone the horn clause with the replacement
                    tx_clause_clone = copy.deepcopy(tx_horn_clause)
#                    print "\tClause after replacement: ", tx_clause_clone
                    for hc in ExtendN3Rules(network,NormalizeClause(tx_clause_clone),constructNetwork):
                        ruleset.append(makeRule(hc,network.nsMap))
                    #restore the replaced term (for the subsequent iteration)
                    list(breadth_first_replace(tx_horn_clause.body,candidate=item,replacement=disj))
            else:
#                print "No Disjunction in the body"
#                print tx_horn_clause
                for hc in ExtendN3Rules(network,NormalizeClause(tx_horn_clause),constructNetwork):
                    _rule=makeRule(hc,network.nsMap)
                    if _rule is not None:
                        ruleset.append(_rule)                    
            #Extract free variables anre add rule to ruleset
#        print "#######################"
#    print "########## Finished Building decision network from DLP ##########"
    #renderNetwork(network).write_graphviz('out.dot')
    return ruleset

def IsaFactFormingConclusion(head):
    """
    'Relative to the def-Horn ruleset, the def-LP is thus sound; moreover, it is 
    complete for fact-form conclusions, i.e., for queries whose answers amount 
    to conjunctions of facts. However, the def-LP is a mildly weaker version of 
    the def-Horn ruleset, in the following sense. Every conclusion of the def-LP
    must have the form of a fact. By contrast, the entailments, i.e., conclusions, 
    of the def-Horn ruleset are not restricted to be facts.' - Scan depth-first
    looking for Clauses
    """
    if isinstance(head,And):
        for i in head:
            if not IsaFactFormingConclusion(i):
                return False
        return True
    elif isinstance(head,Or):
        return False
    elif isinstance(head,Atomic):
        return True
    elif isinstance(head,OriginalClause):
        return False
    else:
        print head
        raise

def traverseClause(condition):
    if isinstance(condition,SetOperator):
        for i in iter(condition):
            yield i
    elif isinstance(condition,Atomic):
        return 

def breadth_first(condition,children=traverseClause):
    """Traverse the nodes of a tree in breadth-first order.
    The first argument should be the tree root; children
    should be a function taking as argument a tree node and
    returning an iterator of the node's children.
    
    From http://ndirty.cute.fi/~karttu/matikka/Python/eppsteins_bf_traversal_231503.htm
    
    """
    yield condition
    last = condition
    for node in breadth_first(condition,children):
        for child in children(node):
            yield child
            last = child
        if last == node:
            return

def breadth_first_replace(condition,
                          children=traverseClause,
                          candidate=None,
                          replacement=None):
    """Traverse the nodes of a tree in breadth-first order.
    The first argument should be the tree root; children
    should be a function taking as argument a tree node and
    returning an iterator of the node's children.
    
    From http://ndirty.cute.fi/~karttu/matikka/Python/eppsteins_bf_traversal_231503.htm
    
    """
    yield condition
    last = condition
    for node in breadth_first_replace(condition,
                                      children,
                                      candidate,
                                      replacement):
        for child in children(node):
            yield child
            if candidate and child is candidate:
                #replace candidate with replacement
                i=node.formulae.index(child)
                node.formulae[i]=replacement
                return
            last = child
        if last == node:
            return

def ExtendN3Rules(network,horn_clause,constructNetwork=False):
    """
    Extends the network with the given Horn clause (rule)
    """
    rt=[]
    if constructNetwork:
        ruleStore = network.ruleStore
        lhs = BNode()
        rhs = BNode()
    assert isinstance(horn_clause.body,(And,Uniterm)),list(horn_clause.body)
    assert len(list(horn_clause.body))
#    print horn_clause
    if constructNetwork:
        for term in horn_clause.body:
            ruleStore.formulae.setdefault(lhs,Formula(lhs)).append(term.toRDFTuple())
    assert isinstance(horn_clause.head,(And,Uniterm))

    if IsaFactFormingConclusion(horn_clause.head):
        def extractVariables(term,existential=True):
            if isinstance(term,existential and BNode or Variable):
                yield term
            elif isinstance(term,Uniterm):
                for t in term.toRDFTuple():
                    if isinstance(t,existential and BNode or Variable):
                        yield t
                        
        def iterCondition(condition):
            return isinstance(condition,SetOperator) and condition or iter([condition])
                
        #first we identify body variables                        
        bodyVars = set(reduce(lambda x,y:x+y,
                              [ list(extractVariables(i,existential=False)) for i in iterCondition(horn_clause.body) ]))
        
        #then we identify head variables
        headVars = set(reduce(lambda x,y:x+y,
                              [ list(extractVariables(i,existential=False)) for i in iterCondition(horn_clause.head) ]))
        
        #then we identify those variables that should (or should not) be converted to skolem terms
        updateDict       = dict([(var,BNode()) for var in headVars if var not in bodyVars])
        
        for uniTerm in iterCondition(horn_clause.head):
            newArg      = [ updateDict.get(i,i) for i in uniTerm.arg ]
            uniTerm.arg = newArg
                                
        exist=[list(extractVariables(i)) for i in breadth_first(horn_clause.head)]
        e=Exists(formula=horn_clause.head,
                 declare=set(reduce(lambda x,y:x+y,exist,[])))        
        if reduce(lambda x,y:x+y,exist):
            horn_clause.head=e
            assert e.declare,exist
        if constructNetwork:
            for term in horn_clause.head:
                assert not hasattr(term,'next')
                if isinstance(term,Or):
                    ruleStore.formulae.setdefault(rhs,Formula(rhs)).append(term)
                else:
                    ruleStore.formulae.setdefault(rhs,Formula(rhs)).append(term.toRDFTuple())
            ruleStore.rules.append((ruleStore.formulae[lhs],ruleStore.formulae[rhs]))
            network.buildNetwork(iter(ruleStore.formulae[lhs]),
                                 iter(ruleStore.formulae[rhs]),
                                 horn_clause)
            network.alphaNodes = [node for node in network.nodes.values() if isinstance(node,AlphaNode)]
        rt.append(horn_clause)
    else:
        for hC in LloydToporTransformation(horn_clause,fullReduction=True):
            rt.append(hC)
            #print "normalized clause: ", hC
            for i in ExtendN3Rules(network,hC,constructNetwork):
                rt.append(hC)
    return rt

def generatorFlattener(gen):
    assert hasattr(gen,'next')
    i = list(gen)
    i = len(i)>1 and [hasattr(i2,'next') and generatorFlattener(i2) or i2 for i2 in i] or i[0]
    if hasattr(i,'next'):
        i=listOrThingGenerator(i)
        #print i
        return i
    elif isinstance(i,SetOperator):
        i.formulae = [hasattr(i2,'next') and generatorFlattener(i2) or i2 for i2 in i.formulae]
        #print i
        return i
    else:
        return i

def SkolemizeExistentialClasses(term,check=False):
    if check:
        return isinstance(term,BNode) and SKOLEMIZED_CLASS_NS[term] or term
    return SKOLEMIZED_CLASS_NS[term]

def NormalizeBooleanClassOperand(term,owlGraph):
    return ((isinstance(term,BNode) and IsaBooleanClassDescription(term,owlGraph)) or \
             IsaRestriction(term,owlGraph))\
          and SkolemizeExistentialClasses(term) or term    

def IsaBooleanClassDescription(term,owlGraph):
    for s,p,o in owlGraph.triples_choices((term,[OWL_NS.unionOf,
                                                OWL_NS.intersectionOf],None)):
        return True

def IsaRestriction(term,owlGraph):
    return (term,RDF.type,OWL_NS.Restriction) in owlGraph
    
def T(owlGraph,complementExpansions=[]):
    """
    #Subsumption (purely for TBOX classification)
    {?C rdfs:subClassOf ?SC. ?A rdfs:subClassOf ?C} => {?A rdfs:subClassOf ?SC}.
    {?C owl:equivalentClass ?A} => {?C rdfs:subClassOf ?A. ?A rdfs:subClassOf ?C}.
    {?C rdfs:subClassOf ?SC. ?SC rdfs:subClassOf ?C} => {?C owl:equivalentClass ?SC}.
    
    T(rdfs:subClassOf(C,D))       -> Th(D(y)) :- Tb(C(y))
    
    T(owl:equivalentClass(C,D)) -> { T(rdfs:subClassOf(C,D) 
                                     T(rdfs:subClassOf(D,C) }
    
    A generator over the Logic Programming rules which correspond
    to the DL  ( unary predicate logic ) subsumption axiom described via rdfs:subClassOf
    """
    for c,p,d in owlGraph.triples((None,RDFS.subClassOf,None)):
        yield NormalizeClause(Clause(Tb(owlGraph,c),Th(owlGraph,d)))
        assert isinstance(c,URIRef),"%s is a kind of %s"%(c,d)
    for c,p,d in owlGraph.triples((None,OWL_NS.equivalentClass,None)):
        yield NormalizeClause(Clause(Tb(owlGraph,c),Th(owlGraph,d)))
        yield NormalizeClause(Clause(Tb(owlGraph,d),Th(owlGraph,c)))
    for s,p,o in owlGraph.triples((None,OWL_NS.intersectionOf,None)):
        if s not in complementExpansions:
            conjunction=[]
            for bodyTerm in Collection(owlGraph,o):
                normalizedBodyTerm=NormalizeBooleanClassOperand(bodyTerm,owlGraph)
                bodyUniTerm = Uniterm(RDF.type,[Variable("X"),normalizedBodyTerm],
                                      newNss=owlGraph.namespaces())                                    
                classifyingClause = NormalizeClause(Clause(Tb(owlGraph,bodyTerm),
                                                 bodyUniTerm))
                if isinstance(bodyTerm,URIRef) and bodyTerm.find(SKOLEMIZED_CLASS_NS)==-1:
                    conjunction.append(bodyUniTerm)
                elif (bodyTerm,OWL_NS.someValuesFrom,None) in owlGraph or\
                     (bodyTerm,OWL_NS.hasValue,None) in owlGraph:                    
                    conjunction.extend(NormalizeClause(Clause(Tb(owlGraph,bodyTerm),None)).body)
                elif (bodyTerm,OWL_NS.allValuesFrom,None) in owlGraph:
                    conjunction.append(bodyUniTerm)                    
                    yield classifyingClause
                elif (bodyTerm,OWL_NS.hasValue,None) in owlGraph:
                    conjunction.extend(NormalizeClause(Clause(Tb(owlGraph,bodyTerm),None)).body)
                elif (bodyTerm,OWL_NS.unionOf,None) in owlGraph:
                    conjunction.append(bodyUniTerm)                    
                    yield classifyingClause
                elif (bodyTerm,OWL_NS.intersectionOf,None) in owlGraph:
                    conjunction.append(bodyUniTerm)                    
                    
            body = And(conjunction)
            head = Uniterm(RDF.type,[Variable("X"),
                                     SkolemizeExistentialClasses(s,check=True)],
                                     newNss=owlGraph.namespaces())
#            O1 ^ O2 ^ ... ^ On => S(?X)            
            yield Clause(body,head)
            if isinstance(s,URIRef):
#                S(?X) => O1 ^ O2 ^ ... ^ On                
    #            special case, owl:intersectionOf is a neccessary and sufficient
    #            criteria and should thus work in *both* directions 
    #            This rule is not added for anonymous classes
                yield Clause(head,body)
        
    for s,p,o in owlGraph.triples((None,OWL_NS.unionOf,None)):
        if isinstance(s,URIRef):
            #special case, owl:unionOf is a neccessary and sufficient
            #criteria and should thus work in *both* directions
            body = Or([Uniterm(RDF.type,[Variable("X"),
                                         NormalizeBooleanClassOperand(i,owlGraph)],
                                         newNss=owlGraph.namespaces()) \
                           for i in Collection(owlGraph,o)])
            head = Uniterm(RDF.type,[Variable("X"),s],newNss=owlGraph.namespaces())
            yield Clause(body,head)
    for s,p,o in owlGraph.triples((None,OWL_NS.inverseOf,None)):
        #    T(owl:inverseOf(P,Q))          -> { Q(x,y) :- P(y,x)
        #                                        P(y,x) :- Q(x,y) }
        newVar = Variable(BNode())
        body1 = Uniterm(s,[newVar,Variable("X")],newNss=owlGraph.namespaces())
        head1 = Uniterm(o,[Variable("X"),newVar],newNss=owlGraph.namespaces())
        yield Clause(body1,head1)
        newVar = Variable(BNode())
        body2 = Uniterm(o,[Variable("X"),newVar],newNss=owlGraph.namespaces())
        head2 = Uniterm(s,[newVar,Variable("X")],newNss=owlGraph.namespaces())
        yield Clause(body2,head2)
    for s,p,o in owlGraph.triples((None,RDF.type,OWL_NS.TransitiveProperty)):
        #T(owl:TransitiveProperty(P))   -> P(x,z) :- P(x,y) ^ P(y,z)
        y = Variable(BNode())
        z = Variable(BNode())
        x = Variable("X")
        body = And([Uniterm(s,[x,y],newNss=owlGraph.namespaces()),\
                    Uniterm(s,[y,z],newNss=owlGraph.namespaces())])
        head = Uniterm(s,[x,z],newNss=owlGraph.namespaces())
        yield Clause(body,head)

    #Contribution (Symmetric DL roles)
    for s,p,o in owlGraph.triples((None,RDF.type,OWL_NS.SymmetricProperty)):
        #T(owl:SymmetricProperty(P))   -> P(y,x) :- P(x,y)
        y = Variable("Y")
        x = Variable("X")
        body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
        head = Uniterm(s,[y,x],newNss=owlGraph.namespaces())
        yield Clause(body,head)
        
    for s,p,o in owlGraph.triples_choices((None,
                                           [RDFS.range,RDFS.domain],
                                           None)):
        if p == RDFS.range:
            #T(rdfs:range(P,D))  -> D(y) := P(x,y)        
            x = Variable("X")
            y = Variable(BNode())
            body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
            head = Uniterm(RDF.type,[y,o],newNss=owlGraph.namespaces())
            yield Clause(body,head)
        else: 
            #T(rdfs:domain(P,D)) -> D(x) := P(x,y)
            x = Variable("X")
            y = Variable(BNode())
            body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
            head = Uniterm(RDF.type,[x,o],newNss=owlGraph.namespaces())
            yield Clause(body,head)
            
def LloydToporTransformation(clause,fullReduction=False):
    """
    Tautological, common horn logic forms (useful for normalizing 
    conjunctive & disjunctive clauses)
    
    (H ^ H0) :- B                 -> { H  :- B
                                       H0 :- B }
    (H :- H0) :- B                -> H :- B ^ H0
    
    H :- (B v B0)                 -> { H :- B
                                       H :- B0 }
    """
    assert isinstance(clause,OriginalClause),repr(clause)
    if isinstance(clause.body,Or):
        for atom in clause.body.formulae:
            yield Clause(atom,clause.head)
    elif isinstance(clause.head,OriginalClause):
        yield Clause(And([clause.body,clause.head.body]),clause.head.head)
    elif isinstance(clause.head,Or):
        #Disjunction in the body, not supported by def-Horn
        #skip
        return
    elif not isinstance(clause.body,Condition):
        print clause
        raise
    elif fullReduction and isinstance(clause.head,And):
        for i in clause.head:
            for j in LloydToporTransformation(Clause(clause.body,i),
                                              fullReduction=fullReduction):
                if [i for i in breadth_first(j.head) if isinstance(i,And)]:
                    #Ands in the head need to be further flattened
                    yield NormalizeClause(j) 
                else:
                    yield j
    else:
        yield clause
    

def commonConjunctionMapping(owlGraph,conjuncts,innerFunc,variable=Variable("X")):
    """
    DHL: T*((C1 ^ C2 ^ ... ^ Cn),x)    -> T*(C1,x) ^ T*(C2,x) ^ ... ^ T*(Cn,x)
    """
    conjuncts = Collection(owlGraph,conjuncts)
    return And([generatorFlattener(innerFunc(owlGraph,c,variable)) 
                   for c in conjuncts])

def Th(owlGraph,_class,variable=Variable('X'),position=LHS):
    """
    DLP head (antecedent) knowledge assertional forms (ABox assertions, conjunction of
    ABox assertions, and universal role restriction assertions)
    """
    props = list(set(owlGraph.predicates(subject=_class)))
    if OWL_NS.allValuesFrom in props:
        #http://www.w3.org/TR/owl-semantics/#owl_allValuesFrom
        #restriction(p allValuesFrom(r))    {x ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† O | <x,y> ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† ER(p) implies y ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† EC(r)}
        for s,p,o in owlGraph.triples((_class,OWL_NS.allValuesFrom,None)):
            prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
            newVar = Variable(BNode())
            body = Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces())
            for head in Th(owlGraph,o,variable=newVar):
                yield Clause(body,head)
    elif OWL_NS.someValuesFrom in props:
        #http://www.w3.org/TR/owl-semantics/#someValuesFrom
        #estriction(p someValuesFrom(e)) {x ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† O | ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ¬¨¬®¬¨¬¢ <x,y> ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† ER(p) ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ¬¨¬®‚Äö√†¬¥ y ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† EC(e)}
        for s,p,o in owlGraph.triples((_class,OWL_NS.someValuesFrom,None)):
            prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
            newVar = BNode()
            yield And([Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces()),
                        generatorFlattener(Th(owlGraph,o,variable=newVar))])
    else:
        #Simple class
        assert not isinstance(_class,BNode),\
        "Cannot conclude existensial set membership"+_class
        yield Uniterm(RDF.type,[variable,_class],newNss=owlGraph.namespaces())
    
def Tb(owlGraph,_class,variable=Variable('X')):
    """
    DLP body (consequent knowledge assertional forms (ABox assertions, 
    conjunction / disjunction of ABox assertions, and exisential role restriction assertions)
    These are all common EL++ templates for KR
    """
    props = list(set(owlGraph.predicates(subject=_class)))
    if OWL_NS.unionOf in props and not isinstance(_class,URIRef):
        #http://www.w3.org/TR/owl-semantics/#owl_unionOf
        #OWL semantics: unionOf(c1 ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂¬¨¬•¬¨¬®¬¨¬Æ¬¨¬®¬¨√Ü‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚àö¬∞ cn) => EC(c1) ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂¬¨¬•¬¨¬®¬¨¬Æ¬¨¬®¬¨¬¢ ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂¬¨¬•¬¨¬®¬¨¬Æ¬¨¬®¬¨√Ü‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚àö¬∞ ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂¬¨¬•¬¨¬®¬¨¬Æ¬¨¬®¬¨¬¢ EC(cn)
        for s,p,o in owlGraph.triples((_class,OWL_NS.unionOf,None)):
            yield Or([Tb(owlGraph,c,variable=variable) \
                           for c in Collection(owlGraph,o)])
    elif OWL_NS.someValuesFrom in props:
        #http://www.w3.org/TR/owl-semantics/#owl_someValuesFrom
        #estriction(p someValuesFrom(e)) {x ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† O | ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ¬¨¬®¬¨¬¢ <x,y> ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† ER(p) ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ¬¨¬®‚Äö√†¬¥ y ‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√†√∂‚àö¬¥‚Äö√Ñ√∂‚àö‚Ä†‚àö‚àÇ‚Äö√Ñ√∂‚àö‚Ä†‚àö√°‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä†‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚Äö√Ñ‚Ä†‚Äö√†√∂‚Äö√†√á‚Äö√Ñ√∂‚àö√ë‚àö‚àÇ‚Äö√†√∂‚àö√´‚Äö√Ñ√∂‚àö√ë‚Äö√Ñ‚Ä† EC(e)}
        prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
        o =list(owlGraph.objects(subject=_class,predicate=OWL_NS.someValuesFrom))[0]
        newVar = Variable(BNode())
        body = Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces())
        head = Th(owlGraph,o,variable=newVar)
        yield And([Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces()),
                    generatorFlattener(Tb(owlGraph,o,variable=newVar))])
    elif OWL_NS.hasValue in props:
        #http://www.w3.org/TR/owl-semantics/#owl_hasValue
        #Domain-specific rules for hasValue
        #Can be achieved via pD semantics        
        prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
        o =first(owlGraph.objects(subject=_class,predicate=OWL_NS.hasValue))
        yield Uniterm(prop,[variable,o],newNss=owlGraph.namespaces())
    else:
        #simple class
        #"Named" Uniterm
        _classTerm=SkolemizeExistentialClasses(_class,check=True)
        yield Uniterm(RDF.type,[variable,_classTerm],newNss=owlGraph.namespaces())


#!/bin/env python
import unittest, os, time, sys
from cStringIO import StringIO
from FuXi.Syntax.InfixOWL import *
from FuXi.Horn import ComplementExpansion
from FuXi.Rete import ReteNetwork
from FuXi.Rete.RuleStore import N3RuleStore
from FuXi.DLP import MapDLPtoNetwork, non_DHL_OWL_Semantics
from FuXi.Rete.Proof import GenerateProof
from FuXi.Rete.Util import generateTokenSet, renderNetwork
from rdflib import URIRef, RDF, RDFS
from rdflib.syntax.NamespaceManager import NamespaceManager
from rdflib.Graph import ConjunctiveGraph, Graph, ReadOnlyGraphAggregate, Namespace

HEART_NS           = Namespace('tag:info@semanticdb.ccf.org,2007:HVICoreDataElements#')
PTREC_NS           = Namespace('tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#')

ONE_OF_AXIOMS_2=\
"""<?xml version="1.0"?>
<!DOCTYPE rdf:RDF [
    <!ENTITY heart "tag:info@semanticdb.ccf.org,2007:HVICoreDataElements#" >
    <!ENTITY sts "tag:info@semanticdb.ccf.org,2008:STS.2.61#" >
    <!ENTITY ptrec "tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#" >
    <!ENTITY dnode "http://www.clevelandclinic.org/heartcenter/ontologies/DataNodes.owl#" >
]>
<rdf:RDF 
     xmlns:dnode="&dnode;"
     xmlns:ptrec="tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#"
     xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
     xmlns:owl="http://www.w3.org/2002/07/owl#"
     xmlns:skos="http://www.w3.org/2004/02/skos/core#"
     xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <owl:Class rdf:about="&ptrec;VADInsertion">
        <skos:editorialNote>Implemented using DLP / n3-horn</skos:editorialNote>
        <owl:intersectionOf rdf:parseType="Collection">
            <rdf:Description rdf:about="&ptrec;SurgicalProcedure_assist_device_assist_device_insertion"/>
            <owl:Restriction>
                <owl:onProperty rdf:resource="&dnode;contains"/>
                <owl:someValuesFrom>
                    <owl:Class>
                        <owl:intersectionOf rdf:parseType="Collection">
                            <owl:Restriction>
                                <owl:onProperty rdf:resource="&dnode;contains"/>
                                <owl:someValuesFrom>
                                    <owl:Class>
                                        <owl:intersectionOf rdf:parseType="Collection">
                                            <rdf:Description rdf:about="&ptrec;CardiacAssistDevice"/>
                                            <owl:Restriction>
                                                <owl:onProperty rdf:resource="&ptrec;hasCardiacAssistDeviceType"/>
                                                <owl:someValuesFrom>
                                                    <owl:Class>
                                                        <owl:oneOf rdf:parseType="Collection">
                                                            <rdf:Description rdf:about="&ptrec;CardiacAssistDeviceType_percutaneous_ventricular_assist"/>
                                                        </owl:oneOf>
                                                    </owl:Class>
                                                </owl:someValuesFrom>
                                            </owl:Restriction>
                                        </owl:intersectionOf>
                                    </owl:Class>
                                </owl:someValuesFrom>
                            </owl:Restriction>
                            <rdf:Description rdf:about="&ptrec;CardiacAssistDeviceInsertion"/>
                        </owl:intersectionOf>
                    </owl:Class>
                </owl:someValuesFrom>
            </owl:Restriction>
        </owl:intersectionOf>
        <rdfs:subClassOf rdf:resource="&sts;VADInsertion"/>
    </owl:Class>
</rdf:RDF>"""

OWL_1=\
"""<?xml version="1.0"?>
<!DOCTYPE rdf:RDF [
    <!ENTITY heart "tag:info@semanticdb.ccf.org,2007:HVICoreDataElements#" >
    <!ENTITY sts "tag:info@semanticdb.ccf.org,2008:STS.2.61#" >
    <!ENTITY ptrec "tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#" >
    <!ENTITY dnode "http://www.clevelandclinic.org/heartcenter/ontologies/DataNodes.owl#" >
    <!ENTITY csqr "tag:info@semanticdb.ccf.org,2008:CardiacSurgeryQualityReport#" >
]>
<rdf:RDF
    xmlns:dnode="&dnode;"
    xmlns:ptrec="tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#"
    xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
    xmlns:owl="http://www.w3.org/2002/07/owl#"
    xmlns:skos="http://www.w3.org/2004/02/skos/core#"
    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <owl:Class rdf:about="&sts;ReoperationForBleeding">
        <owl:intersectionOf rdf:parseType="Collection">
            <owl:Class>
                <owl:intersectionOf rdf:parseType="Collection">
                    <rdf:Description rdf:about="&ptrec;Event_management_operation"/>
                    <owl:Restriction>
                        <owl:onProperty rdf:resource="&dnode;contains"/>
                        <owl:someValuesFrom rdf:resource="&ptrec;SurgicalProcedure_thoracic_control_bleeding"/>
                    </owl:Restriction>
                </owl:intersectionOf>
            </owl:Class>
            <rdf:Description rdf:about="&csqr;PostOpInHospitalEvent"/>
        </owl:intersectionOf>
    </owl:Class> 
</rdf:RDF>"""

OWL_2=\
"""
<!DOCTYPE rdf:RDF [
    <!ENTITY heart "tag:info@semanticdb.ccf.org,2007:HVICoreDataElements#" >
    <!ENTITY sts "tag:info@semanticdb.ccf.org,2008:STS.2.61#" >
    <!ENTITY ptrec "tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#" >
    <!ENTITY dnode "http://www.clevelandclinic.org/heartcenter/ontologies/DataNodes.owl#" >
    <!ENTITY csqr "tag:info@semanticdb.ccf.org,2008:CardiacSurgeryQualityReport#" >
]>
<rdf:RDF
    xmlns:dnode="&dnode;"
    xmlns:ptrec="tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#"
    xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
    xmlns:owl="http://www.w3.org/2002/07/owl#"
    xmlns:skos="http://www.w3.org/2004/02/skos/core#"
    xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <owl:Class rdf:about="&heart;AorticAneurysmRepair">
        <rdfs:subClassOf rdf:resource="&sts;AorticAneurysmRepair"/>
    </owl:Class>    
</rdf:RDF>"""

class OWLDLPTest(unittest.TestCase):
    def setUp(self):
        nsMgr = NamespaceManager(Graph())
        nsMgr.bind(u'ptrec',PTREC_NS)
        self.tBoxGraph = Graph(namespace_manager=nsMgr).parse(StringIO(OWL_1))
        Individual.factoryGraph = self.tBoxGraph 
#        self.factGraph=Graph().parse('40074929.rdf')
        
    def testRuleStore(self):
        network = ReteNetwork(N3RuleStore())
        network.getNsBindings(self.tBoxGraph.namespace_manager)
        NominalRangeTransformer().transform(self.tBoxGraph)
        list(MapDLPtoNetwork(network,self.tBoxGraph))
        for tNode in network.terminalNodes:
            print tNode.clause.n3()
        print network

class NominalRangeTransformer(object):        
    NOMINAL_QUERY=\
    """SELECT ?RESTRICTION ?INTERMEDIATE_CLASS ?NOMINAL ?PROP
       { ?RESTRICTION owl:onProperty ?PROP;
                      owl:someValuesFrom ?INTERMEDIATE_CLASS .  
         ?INTERMEDIATE_CLASS owl:oneOf ?NOMINAL .  } """    
    
    def transform(self,graph):
        """
        Transforms a 'pure' nominal range into a disjunction of value restrictions
        """
        Individual.factoryGraph = graph
        for restriction,intermediateCl,nominal,prop in graph.query(
                                 self.NOMINAL_QUERY,
                                 initNs={u'owl':OWL_NS}):
            nominalCollection=Collection(graph,nominal)
            #purge restriction
            graph.remove((restriction,None,None))
            newConjunct=BooleanClass(restriction,
                                     OWL_NS.unionOf,
                                     [Property(prop)|value|val 
                                                 for val in nominalCollection],
                                     graph)
            #purge nominalization placeholder
            graph.remove((intermediateCl,None,None))
            nominalCollection.clear()
            graph.remove((nominal,None,None))
            print "..Transformed a nominal relation range to a disjunction of value restrictions.."

class NominalRangeTransformTest(unittest.TestCase):
    def setUp(self):
        nsMgr = NamespaceManager(Graph())
        nsMgr.bind(u'ptrec',PTREC_NS)
        self.tBoxGraph = Graph(namespace_manager=nsMgr).parse(StringIO(ONE_OF_AXIOMS_2))
        Individual.factoryGraph = self.tBoxGraph 
        self.factGraph=Graph().parse('40074929.rdf')
        
    def testRuleStore(self):
        network = ReteNetwork(N3RuleStore())
        network.getNsBindings(self.tBoxGraph.namespace_manager)
        NominalRangeTransformer().transform(self.tBoxGraph)
#        print self.tBoxGraph.serialize(format='turtle')
        _class=Class(PTREC_NS.VADInsertion)
#        print _class.__repr__(True)
#        for restr in self.tBoxGraph.subjects(predicate=OWL_NS.hasValue):
#            print "owl:hasValue restriction: %s %s"%(restr,manchesterSyntax(restr,self.tBoxGraph))
        list(MapDLPtoNetwork(network,self.tBoxGraph))
#        renderNetwork(network,nsMap=network.nsMap).write_jpeg('oneOfNetwork.jpg')
        for tNode in network.terminalNodes:
            print tNode.clause.n3()
        network.feedFactsToAdd(generateTokenSet(self.factGraph))
        print network
#        network.reportConflictSet(True)
        vadIns=first(network.inferredFacts.subjects(predicate=RDF.type,object=_class.identifier))
        self.failUnless(vadIns is None, "There should be no ptrec:VADInsertions!")
#        builder,proof=GenerateProof(network,
#                                    (vadIns,RDF.type,_class.identifier))
#        for entry in builder.trace:
#            print entry
#        builder.renderProof(proof).write_jpeg('oneOfProofTree.jpg')

if __name__ == "__main__":
    unittest.main()



from __future__ import generators
import sys
from sets import Set
from rdflib import BNode, RDF, Namespace
from rdflib.store import Store,VALID_STORE, CORRUPTED_STORE, NO_STORE, UNKNOWN
from rdflib.Literal import Literal
from pprint import pprint
from rdflib.syntax.NamespaceManager import NamespaceManager
from rdflib.term_utils import *
from rdflib.Graph import QuotedGraph, Graph
from rdflib.store.REGEXMatching import REGEXTerm, NATIVE_REGEX, PYTHON_REGEX
from cStringIO import StringIO
from BuiltinPredicates import FILTERS
LOG = Namespace("http://www.w3.org/2000/10/swap/log#")
Any = None

RULE_LHS = 0
RULE_RHS = 1

class N3Builtin(object):
    """
    An binary N3 Filter: A built-in which evaluates to a boolean
    """
    def __init__(self,uri,func,argument,result):
        self.uri = uri
        self.argument = argument
        self.result = result
        self.func = func
        self.variables = [arg for arg in [self.argument,self.result] if isinstance(arg,Variable)]
    def toRDFTuple(self):
        return (self.argument,self.uri,self.result)
    def render(self,argument,result):
        return "<%s>(%s,%s)"%(self.uri,argument,result)
    def __iter__(self):
        for f in [self.uri,self.argument,self.result]:
            yield f
    
    def __repr__(self):
        return "<%s>(%s,%s)"%(self.uri,
                              isinstance(self.argument,Variable) and '?%s'%self.argument or self.argument,
                              isinstance(self.result,Variable) and '?%s'%self.result or self.result)
                              
class Formula(object):
    """
    An N3 Formula.  Consists of an (internal) identifier
    and a *list* of triples
    """
    def __init__(self,identifier):
        self.identifier = identifier
        self.triples = []
    def __len__(self):
        return len(self.triples)
    def __repr__(self):
        return "{%s}"%(repr(self.triples))
    def __getitem__(self, key):
        return self.triples[key]
    def __iter__(self):
        for item in self.triples:
            yield item
    def extend(self,other):
        self.triples.extend(other)
    def append(self,other):
        self.triples.append(other)

class Rule(object):
    """
    An N3 Rule.  consists of two formulae associated via log:implies
    """
    def __init__(self,LHS,RHS):
        self.lhs = LHS
        self.rhs = RHS

    def __repr__(self):
        return "{%s} => {%s}"%(self.lhs,self.rhs)

def SetupRuleStore(n3Stream=None,additionalBuiltins=None):
    """
    Sets up a N3RuleStore, a Graph (that uses it as a store, and )
    """
    ruleStore = N3RuleStore(additionalBuiltins=additionalBuiltins)
    nsMgr = NamespaceManager(Graph(ruleStore))
    ruleGraph = Graph(ruleStore,namespace_manager=nsMgr)            
    if n3Stream:
        ruleGraph.parse(n3Stream,format='n3')
    return ruleStore,ruleGraph
        
class N3RuleStore(Store):
    """    
    A specialized Store which maintains order of statements
    and creates N3Filters, Rules, Formula objects, and other facts
    Ensures builtin filters refer to variables that have preceded

    >>> s=N3RuleStore()
    >>> g=Graph(s)
    >>> src = \"\"\"
    ... @prefix str:   <http://www.w3.org/2000/10/swap/string#>.
    ... @prefix math: <http://www.w3.org/2000/10/swap/math#>.
    ... @prefix log:   <http://www.w3.org/2000/10/swap/log#>.
    ... @prefix : <http://metacognition.info/FuXi/test#>.
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    ... @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    ... @prefix owl: <http://www.w3.org/2002/07/owl#>.    
    ... :a a rdfs:Class;
    ...    :prop1 1;
    ...    :prop2 4.
    ... :b a owl:Class;
    ...    :prop1 2;
    ...    :prop2 4,1,5.
    ... (1 2) :relatedTo (3 4).
    ... { ?X a owl:Class. ?X :prop1 ?M. ?X :prop2 ?N. ?N math:equalTo 3 } => { [] :selected (?M ?N) }.\"\"\"    
    >>> g=g.parse(StringIO(src),format='n3')
    >>> s._finalize()
    >>> len([pred for subj,pred,obj in s.facts if pred == u'http://metacognition.info/FuXi/test#relatedTo'])
    1
    >>> len(s.rules)
    1
    >>> print len(s.rules[0][RULE_LHS])
    4
    >>> print len(s.rules[0][RULE_RHS])
    5
    >>> print s.rules[0][RULE_LHS][1]
    (?X, rdflib.URIRef('http://metacognition.info/FuXi/test#prop1'), ?M)
    >>> print s.rules[0][RULE_LHS][-1]
    <http://www.w3.org/2000/10/swap/math#equalTo>(?N,3)

Description Rule Patterns Compilation
    >>> s=N3RuleStore()
    >>> g=Graph(s)
    >>> src = \"\"\"
    ... @prefix math: <http://www.w3.org/2000/10/swap/math#>.
    ... @prefix : <http://metacognition.info/FuXi/test#>.
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    ... @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    ... @prefix owl: <http://www.w3.org/2002/07/owl#>. 
    ... { ?S a [ rdfs:subClassOf ?C ] } => { ?S a ?C }.\"\"\"
    >>> g=g.parse(StringIO(src),format='n3')
    >>> s._finalize()
    >>> assert s.rules
    >>> assert [pattern for pattern in s.rules[0][RULE_LHS] if isinstance(pattern,tuple) and [term for term in pattern if isinstance(term,BNode) ]],repr(s.rules[0][RULE_LHS])    


Test single fact with collection

    >>> s=N3RuleStore()
    >>> g=Graph(s)
    >>> src = \"\"\"
    ... @prefix math: <http://www.w3.org/2000/10/swap/math#>.
    ... @prefix : <http://metacognition.info/FuXi/test#>.
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    ... @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    ... @prefix owl: <http://www.w3.org/2002/07/owl#>.    
    ... (1 2) :relatedTo owl:Class.\"\"\"    
    >>> g=g.parse(StringIO(src),format='n3')
    >>> s._finalize()
    >>> print len(s.facts)
    5

RHS can only include RDF triples

    >>> s=N3RuleStore()
    >>> g=Graph(s)
    >>> src = \"\"\"
    ... @prefix math: <http://www.w3.org/2000/10/swap/math#>.
    ... @prefix : <http://metacognition.info/FuXi/test#>.
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    ... @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    ... @prefix owl: <http://www.w3.org/2002/07/owl#>.    
    ... {} => { 3 math:lessThan 2}.\"\"\"    
    >>> g=g.parse(StringIO(src),format='n3')
    >>> try: 
    ...   s._finalize()
    ... except Exception,e:
    ...   print e
    Rule RHS must only include RDF triples! (<http://www.w3.org/2000/10/swap/math#lessThan>(3,2))

BuiltIn used out of order

    >>> s=N3RuleStore()
    >>> g=Graph(s)
    >>> src = \"\"\"
    ... @prefix math: <http://www.w3.org/2000/10/swap/math#>.
    ... @prefix : <http://metacognition.info/FuXi/test#>.
    ... { ?M math:lessThan ?Z.  ?R :value ?M; :value2 ?Z} => { ?R a :Selected.  }.\"\"\"    
    >>> try: 
    ...   g=g.parse(StringIO(src),format='n3')
    ... except Exception,e:
    ...   print e
    Builtin refers to variables without previous reference! (<http://www.w3.org/2000/10/swap/math#lessThan>(?M,?Z))

    Empty LHS & RHS
    >>> s=N3RuleStore()
    >>> g=Graph(s)
    >>> src = \"\"\"
    ... @prefix math: <http://www.w3.org/2000/10/swap/math#>.
    ... @prefix : <http://metacognition.info/FuXi/test#>.
    ... @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    ... @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    ... @prefix owl: <http://www.w3.org/2002/07/owl#>.    
    ... {} => {rdf:nil :allClasses ?C}.
    ... {?C owl:oneOf ?L. ?X a ?C. ?L :notItem ?X} => {}.\"\"\"    
    >>> g=g.parse(StringIO(src),format='n3')
    >>> len(s.formulae)
    2
    >>> s._finalize()
    >>> len(s.rules[0][0])
    0
    >>> len(s.rules[1][-1])
    0
    """
    context_aware = True
    formula_aware = True

    def __init__(self, identifier=None, additionalBuiltins=None):
        self.formulae = {}
        self.facts = []
        self.rootFormula = None
        self._lists = {}
        self.currentList = None
        self._listBuffer = []
        self.rules = []
        self.referencedVariables = Set()
        self.nsMgr = {u'skolem':URIRef('http://code.google.com/p/python-dlp/wiki/SkolemTerm#')}
        self.filters=FILTERS
        if additionalBuiltins:
            self.filters.update(additionalBuiltins)
        
    def namespace(self,prefix):
        return self.nsMgr.get(prefix)
    
    def bind(self, prefix, namespace, override=True):
        if override or prefix not in self.nsMgr:
            self.nsMgr[prefix]=namespace

    def prefix(self,namespace):
        return dict([(v,k) for 
                       k,v in self.nsMgr.items()]).get(namespace)
        
    def _unrollList(self,l,listName):
        listTriples = []
        lastItemName = None
        for linkItem in l:
            linkName = l.index(linkItem) == 0 and listName or BNode()
            if lastItemName:
                listTriples.append((lastItemName,RDF.rest,linkName))
            listTriples.append((linkName,RDF.first,linkItem))
            lastItemName = linkName
        listTriples.append((lastItemName,RDF.rest,RDF.nil))
        return listTriples 
        
    def _finalize(self):        
        def unrollFunc(left,right):            
            leftListsToUnroll = []
            rightListsToUnroll = []
            if isinstance(left,tuple):
                s,p,o = left
                leftListsToUnroll = [term for term in [s,o] if term in self._lists]
                if leftListsToUnroll:
                    leftListsToUnroll = reduce(lambda x,y:x+y,[self._unrollList(self._lists[l],l) for l in leftListsToUnroll]) 
                left = [left]
            elif isinstance(left,N3Builtin):
                left = [left]
            if isinstance(right,tuple):
                s,p,o = right
                rightListsToUnroll = [term for term in [s,o] if term in self._lists]
                if rightListsToUnroll:
                    rightListsToUnroll = reduce(lambda x,y:x+y,[self._unrollList(self._lists[l],l) for l in rightListsToUnroll])
                right = [right]
            elif isinstance(right,N3Builtin):
                right = [right]                    
            return left +leftListsToUnroll+right+rightListsToUnroll
        if len(self.facts) == 1:
            s,p,o = self.facts[0]
            listsToUnroll = [term for term in [s,o] if term in self._lists]
            if listsToUnroll:
                self.facts.extend(reduce(lambda x,y:x+y,[self._unrollList(self._lists[l],l) for l in listsToUnroll]))            
        elif self.facts:
            self.facts = reduce(unrollFunc,self.facts)
        for formula in self.formulae.values():
            if len(formula) == 1:
                if isinstance(formula[0],tuple):
                    s,p,o = formula[0]
                    listsToUnroll = [term for term in [s,o] if term in self._lists]
                    if listsToUnroll:
                        listTriples = reduce(lambda x,y:x+y,[self._unrollList(self._lists[l],l) for l in listsToUnroll])
                        formula.extend(listTriples)
            elif len(formula):            
                formula.triples = reduce(unrollFunc,[i for i in formula])
        for lhs,rhs in self.rules:
            for item in self.formulae.get(rhs,[]):
                assert isinstance(item,tuple),"Rule RHS must only include RDF triples! (%s)"%item
        self.rules = [(self.formulae.get(lhs,Formula(lhs)),self.formulae.get(rhs,Formula(rhs))) for lhs,rhs in self.rules]
    
    def _checkVariableReferences(self,referencedVariables,terms,funcObj):
        for term in [i for i in terms if isinstance(i,Variable)]:
            if term not in referencedVariables:
                raise Exception("Builtin refers to variables without previous reference! (%s)"%funcObj)
    
    def add(self, (subject, predicate, obj), context=None, quoted=False):
        if predicate == RDF.first and not isinstance(subject,Variable) and not isinstance(object,Variable):
            if not self.currentList:
                self._listBuffer.append(obj)
                self.currentList = subject
            else:
                self._listBuffer.append(obj)
        elif predicate == RDF.rest and not isinstance(subject,Variable) and not isinstance(object,Variable):
            if obj == RDF.nil:
                self._lists[self.currentList] = [item for item in self._listBuffer]
                self._listBuffer = []
                self.currentList = None
        elif not isinstance(context,QuotedGraph):
            if not self.rootFormula:
                self.rootFormula = context.identifier
            if predicate == LOG.implies:
                self.rules.append((subject.identifier,obj.identifier))
            else:                
                self.facts.append((subject,predicate,obj))
        else:
            formula = self.formulae.get(context.identifier,Formula(context.identifier))
            if predicate in FILTERS:
                newFilter = N3Builtin(predicate,FILTERS[predicate](subject,obj),subject,obj)
                self._checkVariableReferences(self.referencedVariables,[subject,obj],newFilter)
                #print newFilter
                formula.append(newFilter)
            else:
                #print "(%s,%s,%s) pattern in %s"%(subject,predicate,obj,context.identifier)
                variables = [arg for arg in [subject,predicate,obj] if isinstance(arg,Variable)]
                self.referencedVariables.union_update(variables)
                formula.append((subject,predicate,obj))
            self.formulae[context.identifier] = formula
                
    def __repr__(self):
        return ""
    
    def __len__(self, context=None):
        return 0
    
    def optimizeRules(self):
        patternDict = {}
        for lhs,rhs in self.rules:
            for pattern in lhs:
                if not isinstance(pattern,N3Builtin):
                    _hashList = [isinstance(term,(Variable,BNode)) and '\t' or term for term in pattern]
                    patternDict.setdefault(reduce(lambda x,y:x+y,_hashList),Set()).add(pattern)
        for key,vals in patternDict.items():
            if len(vals) > 1:
                print "###### Similar Patterns ######"
                for val in vals:
                    print val 
                print "##############################"
    
def test():
    import doctest
    doctest.testmod()

def test2():
    s=N3RuleStore()
    g=Graph(s)
    src = """
    @prefix math: <http://www.w3.org/2000/10/swap/math#>.
    @prefix : <http://metacognition.info/FuXi/test#>.
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    @prefix owl: <http://www.w3.org/2002/07/owl#>.
    :subj :pred obj.     
    {} => { 3 math:lessThan 2}."""    
    g=g.parse(StringIO(src),format='n3')
    s._finalize()

if __name__ == '__main__':
#    pass
    test()
    #test2()
    


import unittest, os, time, sys

N3_PROGRAM=\
"""
@prefix ptrec: <tag:info@semanticdb.ccf.org,2007:PatientRecordTerms#>.
@prefix dnode: <http://www.clevelandclinic.org/heartcenter/ontologies/DataNodes.owl#>.
@prefix owl:   <http://www.w3.org/2002/07/owl#>.
@prefix rdfs:  <http://www.w3.org/2000/01/rdf-schema#>.
@prefix rdf:   <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix csqr:  <tag:info@semanticdb.ccf.org,2008:CardiacSurgeryQualityReport#>.
@prefix str:   <http://www.w3.org/2000/10/swap/string#>.
@prefix log:   <http://www.w3.org/2000/10/swap/log#>.
@prefix cyc:   <http://sw.cyc.com/2006/07/27/cyc#>.
@prefix sts:   <tag:info@semanticdb.ccf.org,2008:STS.2.61#>.

{ ?HOSP a ptrec:Event_encounter_hospitalization; 
        dnode:contains ?HOSP_START_DATE, ?HOSP_STOP_DATE.  
  ?HOSP_START_DATE a ptrec:EventStartDate; ptrec:hasDateTimeMin ?ENCOUNTER_START.
  ?HOSP_STOP_DATE  a ptrec:EventStopDate; ptrec:hasDateTimeMax ?ENCOUNTER_STOP.
  ?EVT_DATE a ptrec:EventStartDate; ptrec:hasDateTimeMin ?EVT_START_MIN .
  ?EVT dnode:contains ?EVT_DATE ; a ?EVT_KIND .
  ?EVT_KIND log:notEqualTo ptrec:Event_encounter_hospitalization .
  ?EVT_START_MIN str:lessThan ?ENCOUNTER_STOP.
  ?EVT_START_MIN str:greaterThanOrEqualTo ?ENCOUNTER_START } => { ?EVT csqr:hasHospitalization ?HOSP } .
  

"""

class NominalRangeTransformTest(unittest.TestCase):
    def setUp(self):
        nsMgr = NamespaceManager(Graph())
        nsMgr.bind(u'ptrec',PTREC_NS)
        self.tBoxGraph = Graph(namespace_manager=nsMgr).parse(StringIO(ONE_OF_AXIOMS_2))
        Individual.factoryGraph = self.tBoxGraph 
        self.factGraph=Graph().parse('40074929.rdf')
        
    def testRuleStore(self):
        network = ReteNetwork(N3RuleStore())
        network.getNsBindings(self.tBoxGraph.namespace_manager)
        NominalRangeTransformer().transform(self.tBoxGraph)
        print self.tBoxGraph.serialize(format='turtle')
        _class=Class(PTREC_NS.VADInsertion)
#        for restr in self.tBoxGraph.subjects(predicate=OWL_NS.hasValue):
#            print "owl:hasValue restriction: %s %s"%(restr,manchesterSyntax(restr,self.tBoxGraph))
        list(MapDLPtoNetwork(network,self.tBoxGraph))
        for tNode in network.terminalNodes:
            print tNode.clause.n3()
        network.feedFactsToAdd(generateTokenSet(self.factGraph))
        print network
        network.reportConflictSet()
        vadIns=first(network.inferredFacts.subjects(predicate=RDF.type,object=_class.identifier))
        builder,proof=GenerateProof(network,
                                    (vadIns,RDF.type,_class.identifier))
#        for entry in builder.trace:
#            print entry
        builder.renderProof(proof).write_jpeg('oneOfProofTree.jpg')
        renderNetwork(network,nsMap=network.nsMap).write_jpeg('oneOfNetwork.jpg')

if __name__ == "__main__":
    unittest.main()



#!/bin/env python
import unittest
from FuXi.Syntax.InfixOWL import *
from FuXi.Rete.Network import ReteNetwork
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.DLP import SKOLEMIZED_CLASS_NS
from rdflib import Namespace
from rdflib.Graph import Graph

EX = Namespace('http://example.com#')

class UnionSkolemizedTest(unittest.TestCase):
    def setUp(self):
        self.tBoxGraph=Graph()
        self.tBoxGraph.namespace_manager.bind('ex',EX)
        self.tBoxGraph.namespace_manager.bind('owl',OWL_NS)
        Individual.factoryGraph = self.tBoxGraph
        self.classB = Class(EX.b)
        self.classE = Class(EX.e)
        self.classF = Class(EX.f) 
        self.classA = BooleanClass(EX.a,
                                   operator=OWL_NS.unionOf,
                                   members=[self.classE,self.classF])
        self.classC = BooleanClass(EX.c,
                                   operator=OWL_NS.unionOf,
                                   members=[self.classA,self.classB])
        self.ruleStore,self.ruleGraph=SetupRuleStore()

    def testUnionSkolemization(self):
        network=ReteNetwork(self.ruleStore)#,nsMap = self.ruleStore.nsMgr)
        p=network.setupDescriptionLogicProgramming(self.tBoxGraph)
        for p in p:
            self.failIf(p.formula.body.arg[-1].find(SKOLEMIZED_CLASS_NS)>-1,
                        "Rule has a skolem term when it shouldn't!: %s"%p)
    
if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
# encoding: utf-8
"""
Implementation of Sideways Information Passing graph (builds it from a given ruleset)
"""

import unittest, os, sys, itertools
from FuXi.Horn.PositiveConditions import *
from FuXi.Horn.HornRules import Ruleset
from FuXi.Rete.RuleStore import SetupRuleStore, N3Builtin
from FuXi.DLP import SKOLEMIZED_CLASS_NS
from rdflib.util import first
from rdflib.Graph import Graph
from rdflib.Collection import Collection
from testMagic import *
from cStringIO import StringIO
from pprint import pprint;
from rdflib import Namespace

MAGIC = Namespace('http://doi.acm.org/10.1145/28659.28689#')

def iterCondition(condition):
    return isinstance(condition,SetOperator) and condition or iter([condition])

class SIPGraphArc(object):
    """
    A sip for r is a labeled graph that satisfies the following conditions:
    1. Each node is either a subset or a member of P(r) or {ph}.
    2. Each arc is of the form N -> q, with label X, where N is a subset of P (r) or {ph}, q is a
    member of P(r), and X is a set of variables, such that
    (i) Each variable of X appears in N.
    (ii) Each member of N is connected to a variable in X.
    (iii) For some argument of q, all its variables appear in X. Further, each variable of X
    appears in an argument of q that satisfies this condition.    
    """
    def __init__(self, left, right, variables, graph=None, headPassing = False):
        self.variables=variables
        self.left = left
        self.right = right
        self.graph = graph is None and Graph() or graph
        self.arc = SKOLEMIZED_CLASS_NS[BNode()]
        self.graph.add((self.arc,RDF.type,MAGIC.SipArc))
        varsCol = Collection(self.graph,BNode())
        [ varsCol.append(i) for i in self.variables ]
        self.graph.add((self.arc,MAGIC.bindings,varsCol.uri))
        if headPassing:
            self.boundHeadPredicate = True
            self.graph.add((self.left,self.arc,self.right))
        else:
            self.boundHeadPredicate = False
            self.graph.add((self.left,self.arc,self.right))
    def __repr__(self):
        """Visual of graph arc"""
        return "%s - (%s) > %s"%(self.left,self.variables,self.right)        
        
def CollectSIPArcVars(left,right):
    """docstring for CollectSIPArcVars"""
    if isinstance(left,list):
        return set(reduce(lambda x,y:x+y,
                          [GetArgs(t) for t in left])).intersection(GetArgs(right))
    else:
        return set(GetArgs(left)).intersection(GetArgs(right))
        
def GetOp(term):
    if isinstance(term,N3Builtin):
        return term.uri
    elif isinstance(term,Uniterm):
        return term.op == RDF.type and term.arg[-1] or term.op
    else:
        print term
        raise term        
        
def GetArgs(term):
    if isinstance(term,N3Builtin):
        return term.argument
    elif isinstance(term,Uniterm):
        return term.arg
    else:
        raise term        
        
def IncomingSIPArcs(sip,pred):
    """docstring for IncomingSIPArcs"""
    for s,p,o in [(s1,p1,o1) for (s1,p1,o1) in sip.triples((None,None,pred)) 
        if (p1,RDF.type,MAGIC.SipArc) in sip]:
        yield Collection(sip,s),Collection(sip,first(sip.objects(p,MAGIC.bindings)))
        
def validSip(sipGraph):
    if not len(sipGraph): return False
    for bindingCol in sipGraph.query("SELECT ?arc { [ a m:BoundHeadPredicate; ?arc ?pred ]. ?arc m:bindings [ rdf:first ?val ] }",
                                     initNs={'m':MAGIC}):
        return True
    return False
        
def BuildNaturalSIP(clause,derivedPreds=None):
    """
    Natural SIP:
    
    Informally, for a rule of a program, a sip represents a
    decision about the order in which the predicates of the rule will be evaluated, and how values
    for variables are passed from predicates to other predicates during evaluation
    
    >>> ruleStore,ruleGraph=SetupRuleStore(StringIO(PROGRAM2))
    >>> ruleStore._finalize()
    >>> fg=Graph().parse(StringIO(PROGRAM2),format='n3')
    >>> rs=Ruleset(n3Rules=ruleGraph.store.rules,nsMapping=ruleGraph.store.nsMgr)
    >>> for rule in rs: print rule
    Forall ?Y ?X ( ex:sg(?X ?Y) :- ex:flat(?X ?Y) )
    Forall ?Y ?Z4 ?X ?Z1 ?Z2 ?Z3 ( ex:sg(?X ?Y) :- And( ex:up(?X ?Z1) ex:sg(?Z1 ?Z2) ex:flat(?Z2 ?Z3) ex:sg(?Z3 ?Z4) ex:down(?Z4 ?Y) ) )
    >>> sip=BuildNaturalSIP(list(rs)[-1])
    >>> for N,x in IncomingSIPArcs(sip,MAGIC.sg): print N.n3(),x.n3()
    ( <http://doi.acm.org/10.1145/28659.28689#up> <http://doi.acm.org/10.1145/28659.28689#sg> <http://doi.acm.org/10.1145/28659.28689#flat> ) ( ?Z3 )
    ( <http://doi.acm.org/10.1145/28659.28689#up> <http://doi.acm.org/10.1145/28659.28689#sg> ) ( ?Z1 )
    
    >>> sip=BuildNaturalSIP(list(rs)[-1],[MAGIC.sg])
    >>> list(sip.query('SELECT ?q {  ?prop a magic:SipArc . [] ?prop ?q . }',initNs={u'magic':MAGIC}))
    [rdflib.URIRef('http://doi.acm.org/10.1145/28659.28689#sg'), rdflib.URIRef('http://doi.acm.org/10.1145/28659.28689#sg')]
    """
    from FuXi.Rete.Util import permu
    assert isinstance(clause.head,Uniterm),"Only one literal in the head!"
    def collectSip(left,right):
        if isinstance(left,list):
            vars=CollectSIPArcVars(left,right)
            leftList=Collection(sipGraph,None)
            assert len(left)==len(set(left))            
            [leftList.append(i) for i in [GetOp(ii) for ii in left]]
            left.append(right)                        
            arc=SIPGraphArc (leftList.uri,GetOp(right),vars,sipGraph)
            return left
        else:
            vars=CollectSIPArcVars(left,right)
            ph=GetOp(left)
            q=GetOp(right)
            arc=SIPGraphArc(ph,q,vars,sipGraph,headPassing=True)
            sipGraph.add((ph,RDF.type,MAGIC.BoundHeadPredicate))
            rt=[left,right]
        return rt
    sipGraph=Graph()  
    if isinstance(clause.body,And):
        foundSipOrder = False
        bodyOrders = permu(clause.body.formulae)
        while not foundSipOrder:
            try:
                bodyOrder=bodyOrders.next()
                reduce(collectSip,
                       itertools.chain(
                              iterCondition(clause.head),
                              iterCondition(And(bodyOrder))))        
                if not validSip(sipGraph):
                    sipGraph.remove((None,None,None))
                else:
                    foundSipOrder=True
                    sipGraph.sipOrder = And(bodyOrder)
            except StopIteration:
                raise Exception("Couldn't find a valid SIP for %s"%clause)
    else:
        reduce(collectSip,itertools.chain(iterCondition(clause.head),
                                          iterCondition(clause.body)))
        sipGraph.sipOrder = clause.body        
    if derivedPreds:
        # We therefore generalize our notation to allow
        # more succint representation of sips, in which only arcs entering 
        # derived predicates are represented.
        arcsToRemove=[]
        collectionsToClear=[]
        for N,prop,q in sipGraph.query(
            'SELECT ?N ?prop ?q {  ?prop a magic:SipArc . ?N ?prop ?q . }',
            initNs={u'magic':MAGIC}):
            if q not in derivedPreds:
                arcsToRemove.extend([(N,prop,q),(prop,None,None)])
                collectionsToClear.append(Collection(sipGraph,N))
                #clear bindings collection as well
                bindingsColBNode=first(sipGraph.objects(prop,MAGIC.bindings))
                collectionsToClear.append(Collection(sipGraph,bindingsColBNode))
        for removeSts in arcsToRemove:
            sipGraph.remove(removeSts)
        for col in collectionsToClear:
            col.clear()
    return sipGraph

def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()


#!/bin/env python
import unittest, os, time, sys, urllib
from pprint import pprint
from cStringIO import StringIO
from FuXi.Rete.Proof import GenerateProof
from FuXi.Syntax.InfixOWL import *
from FuXi.Horn import ComplementExpansion
from FuXi.Rete.Network import ReteNetwork
from FuXi.Rete.RuleStore import N3RuleStore,SetupRuleStore
from FuXi.DLP import MapDLPtoNetwork, non_DHL_OWL_Semantics, makeRule, SKOLEMIZED_CLASS_NS
from FuXi.Rete.Util import generateTokenSet, renderNetwork
from rdflib.util import first
from rdflib.Literal import _XSD_NS
from rdflib.Collection import Collection
from rdflib import URIRef, RDF, RDFS, Namespace, Variable, Literal, URIRef
from rdflib.syntax.NamespaceManager import NamespaceManager
from rdflib.Graph import ConjunctiveGraph, Graph, ReadOnlyGraphAggregate, Namespace

EX = Namespace('http://example.com#')

class DisjunctionOperatorTest(unittest.TestCase):
    def setUp(self):
        self.tBoxGraph=Graph()
        self.tBoxGraph.namespace_manager.bind('ex',EX)
        self.tBoxGraph.namespace_manager.bind('owl',OWL_NS)
        Individual.factoryGraph = self.tBoxGraph
        self.classB = Class(EX.b)
        self.classE = Class(EX.e)
        self.classF = Class(EX.f) 
        self.classA = self.classE | self.classF
        self.classA.identifier = EX.a        
        self.classC = self.classA | self.classB
        self.classC.identifier = EX.c

    def testDisjunction(self):
        self.failUnless((EX.a,None,None) in self.tBoxGraph, "Missing assertions about %s"%EX.a)
        aList=Collection(self.tBoxGraph,EX.a)
        self.assertEquals(len(aList),2)
        for item in [EX.e,EX.f]:
            self.failUnless(item in aList, "Missing %s in %s"%(item,aList.n3()))
        self.failUnless((EX.c,None,None) in self.tBoxGraph, "Missing assertions about %s"%EX.c)
        cList=Collection(self.tBoxGraph,EX.c)
        self.assertEquals(len(cList),2)
        for item in [EX.a,EX.b]:
            self.failUnless(item in aList, "Missing %s in %s"%(item,cList.n3()))

    
if __name__ == '__main__':
    unittest.main()


"""
Utility functions for a Boost Graph Library (BGL) DiGraph via the BGL Python Bindings
"""
from FuXi.Rete.AlphaNode import AlphaNode

try:    
    import boost.graph as bgl
    bglGraph = bgl.Digraph()
except:
    try:
        from pydot import Node,Edge,Dot
    except:
        import warnings
        warnings.warn("Missing pydot library",ImportWarning)        
        #raise NotImplementedError("Boost Graph Library & Python bindings (or pydot) not installed.  See: see: http://www.osl.iu.edu/~dgregor/bgl-python/")
        
from rdflib.Graph import Graph
from rdflib.syntax.NamespaceManager import NamespaceManager
from rdflib import BNode, Namespace, Collection, Variable
from sets import Set

LOG = Namespace("http://www.w3.org/2000/10/swap/log#")

def xcombine(*seqin):
    '''
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/302478
    returns a generator which returns combinations of argument sequences
    for example xcombine((1,2),(3,4)) returns a generator; calling the next()
    method on the generator will return [1,3], [1,4], [2,3], [2,4] and
    StopIteration exception.  This will not create the whole list of 
    combinations in memory at once.
    '''
    def rloop(seqin,comb):
        '''recursive looping function'''
        if seqin:                   # any more sequences to process?
            for item in seqin[0]:
                newcomb=comb+[item]     # add next item to current combination
                # call rloop w/ remaining seqs, newcomb
                for item in rloop(seqin[1:],newcomb):   
                    yield item          # seqs and newcomb
        else:                           # processing last sequence
            yield comb                  # comb finished, add to list
    return rloop(seqin,[])

def permu(xs):
    """
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496819
    "A recursive function to get permutation of a list"
    
    >>> print list(permu([1,2,3]))
    [[1, 2, 3], [1, 3, 2], [2, 1, 3], [2, 3, 1], [3, 1, 2], [3, 2, 1]]
     
    """
    if len(xs) <= 1:
        yield xs
    else:
        for i in range(len(xs)):
            for p in permu(xs[:i] + xs[i + 1:]):
                yield [xs[i]] + p

def generateTokenSet(graph,debugTriples=[],skipImplies=True):
    """
    Takes an rdflib graph and generates a corresponding Set of ReteTokens
    Note implication statements are excluded from the realm of facts by default
    """
    from FuXi.Rete import ReteToken
    rt = Set()    
    def normalizeGraphTerms(term):
        if isinstance(term,Collection.Collection):
            return term.uri
        else:
            return term
    for s,p,o in graph:
        
        if not skipImplies or p != LOG.implies:
            #print s,p,o             
            debug = debugTriples and (s,p,o) in debugTriples
            rt.add(ReteToken((normalizeGraphTerms(s),
                              normalizeGraphTerms(p),
                              normalizeGraphTerms(o)),debug))
    return rt

def generateBGLNode(dot,node,namespace_manager,identifier):
    from FuXi.Rete import ReteNetwork,BetaNode,BuiltInAlphaNode,AlphaNode
    from BetaNode import LEFT_MEMORY, RIGHT_MEMORY, LEFT_UNLINKING
    vertex = Node(identifier)

    shape = 'circle'
    root        = False
    if isinstance(node,ReteNetwork):     
        root        = True
        peripheries = '3'
    elif isinstance(node,BetaNode) and not node.consequent:     
        peripheries = '1'
        if node.aPassThru:
            label = "Pass-thru Beta node\\n"
        elif node.commonVariables:
            label = "Beta node\\n(%s)"%(','.join(["?%s"%i for i in node.commonVariables]))
        else:
            label = "Beta node"

    elif isinstance(node,BetaNode) and node.consequent:     
        #rootMap[vertex] = 'true'
        peripheries = '2'
        stmts = []
        for s,p,o in node.consequent:
            stmts.append(' '.join([str(namespace_manager.normalizeUri(s)),
              str(namespace_manager.normalizeUri(p)),
              str(namespace_manager.normalizeUri(o))]))
              
        rhsVertex = Node(BNode(),
                         label='"'+'\\n'.join(stmts)+'"',
                         shape='plaintext') 
        edge = Edge(vertex,rhsVertex)
        #edge.color = 'red'
        dot.add_edge(edge)              
        dot.add_node(rhsVertex)      
        if node.commonVariables:
            inst=node.network.instanciations.get(node,0)
            label = str("Terminal node\\n(%s)\\n%d instanciations"%(','.join(["?%s"%i for i in node.commonVariables]),inst))
        else:
            label = "Terminal node"
        
    elif isinstance(node,BuiltInAlphaNode):
        peripheries = '1'
        shape = 'plaintext'
        #label = '..Builtin Source..'
        label = repr(node.n3builtin)
        canonicalFunc = namespace_manager.normalizeUri(node.n3builtin.uri)
        canonicalArg1 = namespace_manager.normalizeUri(node.n3builtin.argument)
        canonicalArg2 = namespace_manager.normalizeUri(node.n3builtin.result)
        label = '%s(%s,%s)'%(canonicalFunc,canonicalArg1,canonicalArg2)
        
    elif isinstance(node,AlphaNode):
        peripheries = '1'
        shape = 'plaintext'
#        widthMap[vertex] = '50em'
        label = ' '.join([isinstance(i,BNode) and i.n3() or str(namespace_manager.normalizeUri(i)) 
                           for i in node.triplePattern])    

    vertex.set_shape(shape)
    vertex.set_label('"%s"'%label)
    vertex.set_peripheries(peripheries)
    if root:
        vertex.set_root('true')
    return vertex

def renderNetwork(network,nsMap = {}):
    """
    Takes an instance of a compiled ReteNetwork and a namespace mapping (for constructing QNames
    for rule pattern terms) and returns a BGL Digraph instance representing the Rete network
    #(from which GraphViz diagrams can be generated)
    """
    from FuXi.Rete import BuiltInAlphaNode
    from BetaNode import LEFT_MEMORY, RIGHT_MEMORY, LEFT_UNLINKING
    dot=Dot(graph_type='digraph')
    namespace_manager = NamespaceManager(Graph())
    for prefix,uri in nsMap.items():
        namespace_manager.bind(prefix, uri, override=False)

    visitedNodes = {}
    edges = []
    idx = 0
    for node in network.nodes.values():
        if not node in visitedNodes and not isinstance(node,BuiltInAlphaNode):
            idx += 1
            visitedNodes[node] = generateBGLNode(dot,node,namespace_manager,str(idx))
            dot.add_node(visitedNodes[node])
    nodeIdxs = {}                        
    for node in network.nodes.values():
        for mem in node.descendentMemory:
            if not mem:
                continue
            bNode = mem.successor
        for bNode in node.descendentBetaNodes:
            for otherNode in [bNode.leftNode,bNode.rightNode]:
                if node == otherNode and (node,otherNode) not in edges:
                    for i in [node,bNode]:
                        if i not in visitedNodes:
                            idx += 1
                            nodeIdxs[i] = idx 
                            visitedNodes[i] = generateBGLNode(dot,i,namespace_manager,str(idx))
                            dot.add_node(visitedNodes[i])
                    edge = Edge(visitedNodes[node],visitedNodes[bNode])
                    dot.add_edge(edge)                                        
                    edges.append((node,bNode))
                    
    return dot

def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()    


#!/bin/env python
import unittest, os, time, sys, urllib
from pprint import pprint
from cStringIO import StringIO
from FuXi.Rete.Proof import GenerateProof
from FuXi.Syntax.InfixOWL import *
from FuXi.Horn import ComplementExpansion
from FuXi.Rete.Network import ReteNetwork
from FuXi.Rete.RuleStore import N3RuleStore,SetupRuleStore
from FuXi.DLP import MapDLPtoNetwork, non_DHL_OWL_Semantics, makeRule, SKOLEMIZED_CLASS_NS
from FuXi.Rete.Util import generateTokenSet, renderNetwork
from rdflib.util import first
from rdflib.Literal import _XSD_NS
from rdflib.Collection import Collection
from rdflib import URIRef, RDF, RDFS, Namespace, Variable, Literal, URIRef
from rdflib.syntax.NamespaceManager import NamespaceManager
from rdflib.Graph import ConjunctiveGraph, Graph, ReadOnlyGraphAggregate, Namespace

EX = Namespace('http://example.com#')
DNODE=Namespace('http://www.clevelandclinic.org/heartcenter/ontologies/DataNodes.owl#')

class DisjunctionOperatorTest(unittest.TestCase):
    def setUp(self):
        self.domainModelGraph=Graph().parse('../../../../SemanticDB/OntologyManager/testing/containsAlias.rdf')
        self.closureDeltaGraph=Graph()
        self.ruleStore,self.ruleGraph=SetupRuleStore()
        print self.domainModelGraph.serialize(format='n3')
        self.network = ReteNetwork(self.ruleStore,
                                   inferredTarget = self.closureDeltaGraph,)
                                   #nsMap = self.ruleStore.nsMgr)
        self.network.parseN3Logic(open('../../../../SemanticDB/OntologyManager/ptrecRules.n3'))
        self.network.feedFactsToAdd(generateTokenSet(self.domainModelGraph))

    def test1(self):
        print self.domainModelGraph.serialize(format='n3')
        self.failUnless((DNODE.DataNodeModel,None,None) in self.closureDeltaGraph, "Missing assertions about dnode:DataNodeModel")
    
if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
from pprint import pprint
from sets import Set
from FuXi.Rete.Proof import GenerateProof
from FuXi.Rete import ReteNetwork
from FuXi.Rete.AlphaNode import SUBJECT,PREDICATE,OBJECT,VARIABLE
from FuXi.Rete.BetaNode import PartialInstanciation, LEFT_MEMORY, RIGHT_MEMORY
from FuXi.Rete.RuleStore import N3RuleStore
from FuXi.Rete.Util import renderNetwork,generateTokenSet, xcombine
from FuXi.DLP import MapDLPtoNetwork, non_DHL_OWL_Semantics
from FuXi.Horn import ComplementExpansion
from FuXi.Syntax.InfixOWL import *
from rdflib.Namespace import Namespace
from rdflib import plugin,RDF,RDFS,URIRef,URIRef,Literal,Variable
from rdflib.store import Store
from cStringIO import StringIO
from rdflib.Graph import Graph,ReadOnlyGraphAggregate,ConjunctiveGraph
from rdflib.syntax.NamespaceManager import NamespaceManager
import unittest, time, warnings

RDFLIB_CONNECTION=''
RDFLIB_STORE='IOMemory'

import getopt, sys

def usage():
    print """USAGE: Fuxi [options] factFile1 factFile2 ...
Options:
  --closure                  If present, the inferred triples are serialized 
                             along with the original triples if asked for. Otherwise
                             (the default behavior), only the inferred triples
                             are serialized
                             
  --output=OUT               Determines whether to serialize the inferred triples
                             to STDOUT using the specified RDF syntax ('xml','pretty-xml',
                             'nt','turtle',or 'n3') or to print a summary of the conflict set 
                             (from the RETE network) if the value of this option is
                             'conflict'.  If the DLP mechanism is invoked (via --dlp) then
                             a value of 'rif' will cause the generated ruleset to be rendered
                             in the RIF format.  If the proof generation mechanism is
                             activated then a value of 'pml' will trigger a serialization
                             of the proof in PML.  
                             
  --man-owl                  If present, either the closure (or just the inferred triples) are serialized 
                             using an extension of the manchester OWL syntax
                             with indications for ontology normalization
                             (http://www.cs.man.ac.uk/~rector/papers/rector-modularisation-kcap-2003-distrib.pdf)
  --normalize                Will attempt to determine if the ontology is 'normalized' [Rector, A. 2003]                             
  --help
  --input-format=<FORMAT>    Determines the format of the RDF document(s) which
                             serve as the initial facts for the RETE network.
                             One of 'xml','n3','trix', 'nt', or 'rdfa'.  The default
                             is 'xml'.
  --pDSemantics              Add pD semantics ruleset?                           
  --optimize                 Suggest inefficiencies in the ruleset and exit
                     
  --stdin                    Parse STDIN as an RDF graph to contribute to the
                             initial facts for the RETE network using the 
                             specified format
                             
  --ns=PREFIX=NSURI          Register a namespace binding (QName prefix to a 
                             base URI).  This can be used more than once
                             
  --graphviz-out=<FILE>      A filename to write a graphviz diagram of the RETE
                             network to
  
  --rules=FILE1,FILE2,..     The Notation 3 documents to use as rulesets for the
                             RETE network
                             
  --ruleFacts               Determines whether or not to attempt to parse 
                             initial facts from the rule graph.  Default by default
                             
  --complementExpand         Perform a closed-world expansion of all use of owl:complementOf                             
                              
  --dlp                      This switch turns on Description Logic Programming 
                             (DLP) inference.  In this mode, the input document 
                             is considered an OWL ontology mostly comprised of
                             Description Horn Logic (DHL) axioms. ontology.  An 
                             additional ruleset is included to capture those 
                             semantics outside DHL but which can be expressed in
                             definite Datalog Logic Programming.  The DHL-compiled 
                             ruleset and the extensions are mapped into a RETE-UL 
                             Network for evaluateion.
   --proove                  A N3 string consisting of a single RDF assertion to proove
                             against the rules and facts provided.  Depending on the 
                             --output switch, the proof can be rendered as a Graphviz dot
                             graph, as a PML proof document, or in a human-readable printout                           
"""    
def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "", ["optimize",
                                                      "output=",
                                                      "ns=",
                                                      "proove=",
                                                      "facts=", 
                                                      "rules=",
                                                      "normalize",
                                                      "man-owl",
                                                      "dlp",
                                                      "pDSemantics",
                                                      "complementExpand",
                                                      "stdin",
                                                      "help",
                                                      "ruleFacts",
                                                      "graphviz-out=",
                                                      "input-format=",
                                                      "closure"])
    except getopt.GetoptError, e:
        # print help information and exit:
        print e
        usage()
        sys.exit(2)
    pDSemantics=False
    complementExpansion = False
    proove=None
    factGraphs = args
    ruleGraphs = []
    factFormat = 'xml'
    useRuleFacts = False
    gVizOut = None
    nsBinds = {'iw':'http://inferenceweb.stanford.edu/2004/07/iw.owl#'}
    outMode = 'n3'
    optimize = False
    stdIn = False
    closure = False
    dlp = False
    manOWL = False
    normalize=False
    if not opts:
        usage()
        sys.exit()        
    for o, a in opts:
        if o == '--input-format':
            factFormat = a
        elif o == '--pDSemantics':
            pDSemantics=True
        elif o == '--stdin':
            stdIn = True
        elif o == '--optimize':
            optimize = True            
        elif o == '--normalize':
            normalize = True            
        elif o == '--output':
            outMode = a
        elif o == '--ns':            
            pref,nsUri = a.split('=')
            nsBinds[pref]=nsUri
        elif o == '--graphviz-out':
            gVizOut = a
        elif o == '--man-owl':
            manOWL = True
        elif o == "--help":
            usage()
            sys.exit()
        elif o == '--dlp':
            dlp = True
        elif o == '--complementExpand':
            complementExpansion = True
        elif o == "--rules":
            ruleGraphs = a.split(',')
        elif o == '--ruleFacts':
            useRuleFacts = True
        elif o == '--closure':
            closure = True
        elif o == '--proove':
            proove=a
            
    store = plugin.get(RDFLIB_STORE,Store)()        
    store.open(RDFLIB_CONNECTION)
    
    namespace_manager = NamespaceManager(Graph())
    for prefix,uri in nsBinds.items():
        namespace_manager.bind(prefix, uri, override=False)    
    ruleStore=N3RuleStore()
    nsMgr = NamespaceManager(Graph(ruleStore))
    ruleGraph = Graph(ruleStore,namespace_manager=nsMgr)
    closureDeltaGraph = Graph(store)
    closureDeltaGraph.namespace_manager = namespace_manager
    factGraph = Graph(store) 
    factGraph.namespace_manager = namespace_manager
    for fileN in ruleGraphs:
        print >>sys.stderr,"Parsed %s N3 rules from %s"%(len(ruleGraph.parse(open(fileN),
                                                         format='n3')), 
                                                         fileN)
        if useRuleFacts:
            factGraph.parse(open(fileN),format='n3')
            print >>sys.stderr,"Parsing RDF facts from ", fileN
    assert not ruleGraphs or len(ruleGraph),"Nothing parsed from %s"%(ruleGraphs)
    if optimize:
        ruleStore.optimizeRules()
        sys.exit(1)
    if factGraphs:
        for fileN in factGraphs:
            factGraph.parse(fileN,format=factFormat)
    if stdIn:
        factGraph.parse(sys.stdin,format=factFormat)
    workingMemory = generateTokenSet(factGraph)
    nsBinds.update(ruleStore.nsMgr)
    if dlp:
        if complementExpansion:
            Individual.factoryGraph = factGraph
            def topList(node,g):
                for s in g.subjects(RDF.rest,node):
                    yield s
            for negativeClass in factGraph.subjects(predicate=OWL_NS.complementOf):
                containingList = first(factGraph.subjects(RDF.first,negativeClass))
                prevLink = None
                while containingList:
                    prevLink = containingList
                    containingList = first(factGraph.subjects(RDF.rest,containingList))
                for s,p,o in factGraph.triples_choices((None,
                                                    [OWL_NS.intersectionOf,
                                                     OWL_NS.unionOf],
                                                     prevLink)):
                    _class = Class(s)
        #            print _class.__repr__(True,True)            
                    ComplementExpansion(_class)        
        if pDSemantics:
            ruleGraph.parse(StringIO(non_DHL_OWL_Semantics),format='n3')
        network = ReteNetwork(ruleStore,
                              inferredTarget = closureDeltaGraph,
                              graphVizOutFile = gVizOut,
                              nsMap = nsBinds)
        print >>sys.stderr,"Building DLP ruleset"
        start = time.time()  
        rules=MapDLPtoNetwork(network,factGraph)
        sTime = time.time() - start
        if sTime > 1:
            sTimeStr = "%s seconds"%sTime
        else:
            sTime = sTime * 1000
            sTimeStr = "%s milli seconds"%sTime
        print >>sys.stderr,"Time to map Description Horn Logic axioms to definite Horn clauses and import into network: ",sTimeStr
        print >>sys.stderr,network
        if outMode == 'rif':
            for rule in rules:
                print rule
        elif outMode == 'n3':
            for rule in rules:
                print rule.n3()
    else:
        network = ReteNetwork(ruleStore,
                              inferredTarget = closureDeltaGraph,
                              graphVizOutFile = gVizOut,
                              nsMap = nsBinds)
    start = time.time()  
    network.feedFactsToAdd(workingMemory)
    sTime = time.time() - start
    if sTime > 1:
        sTimeStr = "%s seconds"%sTime
    else:
        sTime = sTime * 1000
        sTimeStr = "%s milli seconds"%sTime
    print >>sys.stderr,"Time to calculate closure on working memory: ",sTimeStr
        
    if outMode == 'conflict':
        tNodeOrder = [tNode for tNode in network.terminalNodes if network.instanciations[tNode]]
        tNodeOrder.sort(key=lambda x:network.instanciations[x],reverse=True)
        for termNode in tNodeOrder:
            lhsF,rhsF = termNode.ruleFormulae
            print >>sys.stderr,termNode
            #print "\t %s => %s"%(lhsF,rhsF)
            print >>sys.stderr,"\t", rhsF
            print >>sys.stderr,"\t\t%s instanciations"%network.instanciations[termNode]
    else:        
        if manOWL:
            cGraph = network.closureGraph(factGraph,readOnly=False)
#            cloneGraph = Graph()
#            cloneGraph += cGraph
            cGraph.namespace_manager = namespace_manager
            Individual.factoryGraph = cGraph
            for p in AllProperties(cGraph):
                print p.identifier
                print repr(p)
            for c in AllClasses(cGraph):#cGraph.subjects(predicate=RDF.type,object=OWL_NS.Class):
                if normalize:
                    if c.isPrimitive():
                        primAnc = [sc for sc in c.subClassOf if sc.isPrimitive()] 
                        if len(primAnc)>1:
                            warnings.warn("Branches of primitive skeleton taxonomy should form trees: %s has %s primitive parents: %s"%(c.qname,
                                                                                                                                        len(primAnc),
                                                                                                                                        primAnc),UserWarning,1)
                        children = [desc for desc in c.subSumpteeIds()]
                        for child in children:
                            for otherChild in [o for o in children if o is not child]:
                                if not otherChild in [c.identifier for c in Class(child).disjointWith]:# and\
                                   #not child in [c.identifier for c in Class(otherChild).disjointWith]:
                                    warnings.warn("Primitive children (of %s) must be mutually disjoint: %s and %s"%(
                                                                                    c.qname,
                                                                                    Class(child).qname,
                                                                                    Class(otherChild).qname),UserWarning,1)
                if not isinstance(c.identifier,BNode):
                    print c.__repr__(True)
        elif closure:
            #FIXME: The code below *should* work
            cGraph = network.closureGraph(factGraph)
            cGraph.namespace_manager = namespace_manager
            print cGraph.serialize(destination=None, format=outMode, base=None)
        elif proove:
            goalGraph=Graph()
            goalGraph.parse(StringIO(proove),format='n3')
            print proove,len(goalGraph)
            assert len(goalGraph),"Empty goal!"
            goal=list(goalGraph)[0]
            builder,proof=GenerateProof(network,goal)
            if outMode == 'dot':
                builder.renderProof(proof).write_graphviz('proof.dot')
            elif outMode == 'pml':
                proofGraph=Graph()
                proofGraph.namespace_manager = namespace_manager
                builder.serialize(proof,proofGraph)
                print proofGraph.serialize(format='pretty-xml')                
            else:
                for step in builder.trace:
                    print step
        elif not outMode =='rif':
            print network.inferredFacts.serialize(destination=None, format=outMode, base=None)
        
    print >> sys.stderr, repr(network)
    store.rollback()
if __name__ == "__main__":
    main()
